from unittest.mock import MagicMock, patch
from marshmallow import Schema, fields as ma
from flask import Flask
from flask_restplus import Resource, Api, fields as fr

# from .utils import unpack_list, unpack_nested
import flask_accepts.utils as utils


def test_unpack_list():
    app = Flask(__name__)
    api = Api(app)
    with patch("flask_accepts.utils.unpack_list", wraps=utils.unpack_list) as mock:
        result = utils.unpack_list(ma.List(ma.Integer()), api=api)

        assert isinstance(result, fr.List)
        assert mock.call_count == 1


def test_unpack_list_of_list():
    app = Flask(__name__)
    api = Api(app)
    with patch(
        "flask_accepts.utils.unpack_list", wraps=utils.unpack_list
    ) as mock, patch.dict("flask_accepts.utils.type_map", {ma.List: mock}):

        result = utils.unpack_list(ma.List(ma.List(ma.Integer())), api=api)

        assert isinstance(result, fr.List)
        assert mock.call_count == 2


def test_unpack_nested_list():
    app = Flask(__name__)
    api = Api(app)
    mock_unpack_list = MagicMock(wraps=utils.unpack_list)
    with patch(
        "flask_accepts.utils.unpack_nested", wraps=utils.unpack_nested
    ) as mock_unpack_nested, patch.dict(
        "flask_accepts.utils.type_map", {ma.List: mock_unpack_list}
    ):
        result = utils.unpack_nested(ma.Nested(ma.List(ma.Integer())), api=api)

        assert isinstance(result, fr.Nested)
        assert mock_unpack_nested.call_count == 1
        assert mock_unpack_list.call_count == 1


def test_unpack_nested():
    app = Flask(__name__)
    api = Api(app)
    result = utils.unpack_nested(ma.Nested(ma.Integer()), api=api)

    assert result


def test_get_default_model_name():
    from .utils import get_default_model_name

    class TestSchema(Schema):
        pass

    result = get_default_model_name(TestSchema)

    expected = "Test"
    assert result == expected


def test_get_default_model_name_works_with_multiple_schema_in_name():
    from .utils import get_default_model_name

    class TestSchemaSchema(Schema):
        pass

    result = get_default_model_name(TestSchemaSchema)

    expected = "TestSchema"
    assert result == expected


def test_get_default_model_name_that_does_not_end_in_schema():
    from .utils import get_default_model_name

    class SomeOtherName(Schema):
        pass

    result = get_default_model_name(SomeOtherName)

    expected = "SomeOtherName"
    assert result == expected


def test_get_default_model_name_default_names():
    from .utils import get_default_model_name, num_default_models

    for model_num in range(5):
        result = get_default_model_name()
        expected = f"DefaultResponseModel_{model_num + num_default_models}"
        assert result == expected



from typing import Optional, Type, Union
from flask_restplus import fields as fr
from marshmallow import fields as ma
from marshmallow.schema import Schema, SchemaMeta
import uuid


def unpack_list(val, api, model_name: str = None):
    model_name = model_name = get_default_model_name()
    return fr.List(map_type(val.inner, api, model_name))


def unpack_nested(val, api, model_name: str = None):
    model_name = model_name = get_default_model_name()
    return fr.Nested(map_type(val.nested, api, model_name))


def for_swagger(schema, api, model_name: str = None):
    """
    Convert a marshmallow schema to equivalent Flask-RESTplus model

    Args:
        schema (Marshmallow Schema): Schema defining the inputs
        api (Namespace): Flask-RESTplus namespace (necessary for context)
        model_name (str): Name of Flask-RESTplus model

    Returns:
        api.model: An equivalent api.model
    """

    model_name = model_name or get_default_model_name()

    # For nested Schemas, the internal fields are stored in _declared_fields, whereas
    # for Schemas the name is declared_fields, so check for both.
    if isinstance(schema, SchemaMeta):
        schema = schema()
    fields = {
        k: map_type(v, api, model_name)
        for k, v in (vars(schema).get("fields").items())
        if type(v) in type_map
    }
    return api.model(model_name, fields)


type_map = {
    ma.AwareDateTime: lambda val, api, model_name: fr.Raw(example=val.default),
    ma.Bool: lambda val, api, model_name: fr.Boolean(example=val.default),
    ma.Boolean: lambda val, api, model_name: fr.Boolean(example=val.default),
    ma.Constant: lambda val, api, model_name: fr.Raw(example=val.default),
    ma.Date: lambda val, api, model_name: fr.Date(example=val.default),
    ma.DateTime: lambda val, api, model_name: fr.DateTime(example=val.default),
    # For some reason, fr.Decimal has no example parameter, so use Float instead
    ma.Decimal: lambda val, api, model_name: fr.Float(example=val.default),
    ma.Dict: lambda val, api, model_name: fr.Raw(example=val.default),
    ma.Email: lambda val, api, model_name: fr.String(example=val.default),
    ma.Float: lambda val, api, model_name: fr.Float(example=val.default),
    ma.Function: lambda val, api, model_name: fr.Raw(example=val.default),
    ma.Int: lambda val, api, model_name: fr.Integer(example=val.default),
    ma.Integer: lambda val, api, model_name: fr.Integer(example=val.default),
    ma.Length: lambda val, api, model_name: fr.Float(example=val.default),
    ma.Mapping: lambda val, api, model_name: fr.Raw(example=val.default),
    ma.NaiveDateTime: lambda val, api, model_name: fr.DateTime(example=val.default),
    ma.Number: lambda val, api, model_name: fr.Float(example=val.default),
    ma.Pluck: lambda val, api, model_name: fr.Raw(example=val.default),
    ma.Raw: lambda val, api, model_name: fr.Raw(example=val.default),
    ma.Str: lambda val, api, model_name: fr.String(example=val.default),
    ma.String: lambda val, api, model_name: fr.String(example=val.default),
    ma.Time: lambda val, api, model_name: fr.DateTime(example=val.default),
    ma.Url: lambda val, api, model_name: fr.Url(example=val.default),
    ma.URL: lambda val, api, model_name: fr.Url(example=val.default),
    ma.UUID: lambda val, api, model_name: fr.String(example=val.default),
    ma.List: unpack_list,
    ma.Nested: unpack_nested,
    Schema: for_swagger,
    SchemaMeta: for_swagger,
}

num_default_models = 0


def get_default_model_name(schema: Optional[Union[Schema, Type[Schema]]] = None) -> str:
    if schema:
        if isinstance(schema, Schema):
            return "".join(schema.__class__.__name__.rsplit("Schema", 1))
        else:
            # It is a type itself
            return "".join(schema.__name__.rsplit("Schema", 1))

    global num_default_models
    name = f"DefaultResponseModel_{num_default_models}"
    num_default_models += 1
    return name


def map_type(val, api, model_name):
    return type_map[type(val)](val, api, model_name)
